// (C) 2001-2024 Intel Corporation. All rights reserved.
// Your use of Intel Corporation's design tools, logic functions and other 
// software and tools, and its AMPP partner logic functions, and any output 
// files from any of the foregoing (including device programming or simulation 
// files), and any associated documentation or information are expressly subject 
// to the terms and conditions of the Intel Program License Subscription 
// Agreement, Intel FPGA IP License Agreement, or other applicable 
// license agreement, including, without limitation, that your use is for the 
// sole purpose of programming logic devices manufactured by Intel and sold by 
// Intel or its authorized distributors.  Please refer to the applicable 
// agreement for further details.


// (C) 2001-2023 Intel Corporation. All rights reserved.
// Your use of Intel Corporation's design tools, logic functions and other 
// software and tools, and its AMPP partner logic functions, and any output 
// files from any of the foregoing (including device programming or simulation 
// files), and any associated documentation or information are expressly subject 
// to the terms and conditions of the Intel Program License Subscription 
// Agreement, Intel FPGA IP License Agreement, or other applicable 
// license agreement, including, without limitation, that your use is for the 
// sole purpose of programming logic devices manufactured by Intel and sold by 
// Intel or its authorized distributors.  Please refer to the applicable 
// agreement for further details.


/*
  Legal Notice: (C)2020 Intel Corporation. All rights reserved.  Your
  use of Intel Corporation's design tools, logic functions and other
  software and tools, and its AMPP partner logic functions, and any
  output files any of the foregoing (including device programming or
  simulation files), and any associated documentation or information are
  expressly subject to the terms and conditions of the Intel Program
  License Subscription Agreement or other applicable license agreement,
  including, without limitation, that your use is for the sole purpose
  of programming logic devices manufactured by Intel and sold by Intel
  or its authorized distributors.  Please refer to the applicable
  agreement for further details.
*/

/*
PRBS checker core

Description:  This component checks the input data against the PRBS pattern which is based on the data width chosen or a custom polynomial of programmed during
              runtime.  When the host writes to the start bit of the CSR slave port the component will begin receiving data from the sink port.  You should use
              the prbs generator component with this component.

              The following polynomials are used by default, you may overwrite them by writing the PRBS Polynomial at run time.  The following polynomials
              are used:

              Data width = 8   bits --> PRBS-8   (taps [8,   6,   5,   4]  )
              Data width = 16  bits --> PRBS-16  (taps [16,  14,  13,  11] )
              Data width = 32  bits --> PRBS-32  (taps [32,  30,  26,  25] )
              Data width = 64  bits --> PRBS-64  (taps [64,  63,  61,  60] )
              Data width = 128 bits --> PRBS-128 (taps [128, 127, 126, 121])

              The PRBS polynomials do not include the implied '1' so if you provide your own custom polynomial do not program the implied '1'.  For example
              the PRBS-128 polynomial of x^128 + x^127 + x^126 + x^121 + 1 programs bits 127, 126, 125, and 120 high and the rest are low.  The generator
              is seeded a hardcoded value based on the data width.

Register map:

|-------------------------------------------------------------------------------------------------------------------------------------------------------|
|  Address   |   Access Type   |                                                            Bits                                                        |
|            |                 |------------------------------------------------------------------------------------------------------------------------|
|            |                 |            31..24            |            23..16            |            15..8            |            7..0            |
|-------------------------------------------------------------------------------------------------------------------------------------------------------|
|     0      |       r/w       |                                                       Payload Length                                                   |
|     4      |       N/A       |                                                            N/A                                                         |
|     8      |       r/w       |                           Run                   Seed Checker               Stop on Failure     Infinite Payload Length |
|     12     |       ro        |                                                                                                       Failure Detected |
|     16     |       r/w       |                                                   PRBS Polynomial [31..0]                                              |
|     20     |       r/w       |                                                   PRBS Polynomial [63..32]                                             |
|     24     |       r/w       |                                                   PRBS Polynomial [95..64]                                             |
|     28     |       r/w       |                                                   PRBS Polynomial [127..96]                                            |
|     32     |       r/w       |                                                                                                     Error Counter Clear|
|     36     |       N/A       |                                                            N/A                                                         |
|     40     |       N/A       |                                                            N/A                                                         |
|     44     |       N/A       |                                                            N/A                                                         |
|     48     |       N/A       |                                                            N/A                                                         |
|     52     |       N/A       |                                                            N/A                                                         |
|     56     |       N/A       |                                                            N/A                                                         |
|     60     |       N/A       |                                                            N/A                                                         |
|-------------------------------------------------------------------------------------------------------------------------------------------------------|

Address 0  --> Bits 31:0 are used store the payload length as well as read it back while the checker core is operating.  This field should only be written to while
               the checker is stopped.
Address 4  --> <reserved>
Address 8  --> Bit 0 is used tell the checker to ignore the payload length field and generate the pattern until stopped.  Bit 8 is used to stop the checker when a failure
               is detected, bit 16 is used to seed the checker core and bit 24 is used to start the checker core so that it begins receiving data.  The run field must be
               accessed at the same time as the other two bits or later.
Address 12 --> Bit 0 is used to read back the failure status as well as clear the failure bit
Address 16 --> PRBS polynomial bits 31-0 (does not include the implied '1' in the polynomial)
Address 20 --> PRBS polynomial bits 63-32
Address 24 --> PRBS polynomial bits 95-64
Address 28 --> PRBS polynomial bits 127-96
Address 32 --> Bit 0 is used for error counter clearance, other bits are reserved
Address 36 --> <reserved>
Address 40 --> <reserved>
Address 44 --> <reserved>
Address 48 --> <reserved>
Address 52 --> <reserved>
Address 56 --> <reserved>
Address 60 --> <reserved>

Revision Info:
Original author: JCJB
Initial release 1.0: 04/21/2010

Updated release 1.1: 12/16/2020, Platform Team/CEG/PSG
*/

// synthesis translate_off
`timescale 1ns / 1ps
// synthesis translate_on

// turn off superfluous verilog processor warnings 
// Intel message_level Level1 
// Intel message_off 10230

module  mtm_prbs_pattern_checker        
        #(
        parameter                    DATA_WIDTH = 32,  // must be 8, 16, 32, 64, 128, 256, 512, 1024
        parameter                    PRBS_WIDTH = 16   // must be 8, 16, 32, 64, or 128
        )
        (
        input  wire                  clk,
        input  wire                  reset,

        input  wire [4:0]            csr_address,
        input  wire [31:0]           csr_writedata,
        input  wire                  csr_write,
        input  wire [3:0]            csr_byteenable,
        input  wire                  csr_read,
        output reg  [31:0]           csr_readdata,

        input  wire [DATA_WIDTH-1:0] snk_data,
        input  wire                  snk_valid,
        output wire                  snk_ready
        );

//#######################################################################################
//Local parameter definition
//#######################################################################################
localparam NUM_OF_SYMBOLS         = DATA_WIDTH / 8; // in bytes

//Do not include the implied '1'
localparam POLYNOMIAL_RESET_VALUE = (PRBS_WIDTH == 8)  ? 8'hB8 :                                       // taps [8, 6, 5, 4]
                                    (PRBS_WIDTH == 16) ? 16'hB400 :                                    // taps [16, 14, 13, 11]
                                    (PRBS_WIDTH == 32) ? 32'hA300_0000 :                               // taps [32, 30, 26, 25]
                                    (PRBS_WIDTH == 64) ? 64'hD800_0000_0000_0000 :                     // taps [64, 63, 61, 60]
                                   /*PRBS_WIDTH == 128*/ 128'hE100_0000_0000_0000_0000_0000_0000_0000; // taps [128, 127, 126, 121]

localparam SEED_VALUE             = (PRBS_WIDTH == 8)  ? 8'h5A :
                                    (PRBS_WIDTH == 16) ? 16'hFACE :
                                    (PRBS_WIDTH == 32) ? 32'hDEAD_BEEF :
                                    (PRBS_WIDTH == 64) ? 64'hFEED_ABBA_BABE_BEEF :
                                   /*PRBS_WIDTH == 128*/ 128'hCA11_DEAF_ABBA_BABE_B01D_5CA1_AB1E_FACE;

localparam PATTERN_WALKING_0      = (PRBS_WIDTH == 8)  ? 8'hFE : 
                                    (PRBS_WIDTH == 16) ? 16'hFFFE : 
                                    (PRBS_WIDTH == 32) ? 32'hFFFF_FFFE : 
                                    (PRBS_WIDTH == 64) ? 64'hFFFF_FFFF_FFFF_FFFE : 
                                   /*PRBS_WIDTH == 128*/ 128'hFFFF_FFFF_FFFF_FFFF_FFFF_FFFF_FFFF_FFFE;

localparam PATTERN_WALKING_1      = (PRBS_WIDTH == 8)  ? 8'h01 : 
                                    (PRBS_WIDTH == 16) ? 16'h0001 : 
                                    (PRBS_WIDTH == 32) ? 32'h0000_0001 : 
                                    (PRBS_WIDTH == 64) ? 64'h0000_0000_0000_0001 : 
                                   /*PRBS_WIDTH == 128*/ 128'h0000_0000_0000_0000_0000_0000_0000_0001;

localparam TEST_CONSTANT_DEFAULT  = 128'h5FAF5FAF5FAF5FAF5FAF5FAF5FAF5FAF;

//#######################################################################################
//Interal registers and wires
//#######################################################################################
reg  [63:0]           payload_length_counter;
reg                   infinite_payload_length;
reg                   seed_checker;
reg                   stop_on_failure;
reg                   start;
reg                   start_d1;
reg  [127:0]          test_constant;
reg  [3:0]            test_pattern; // 4'b0000: user defined prbs pattern (64-bit width), default test pattern
                                    // 4'b0001: user defined constant pattern
                                    // 4'b0010: walking '0' started from "FFFF_FFFF_FFFF_FFFE"
                                    // 4'b0011: walking '1' started from "0000_0000_0000_0001"
                                    // Others:  reserved

wire                  clear_start/* synthesis keep */;
reg  [31:0]           error_counter;
reg                   error_counter_clr;
reg                   failure_detected;
wire                  failure_detected_set;
reg  [127:0]          polynomial;
reg  [127:0]          seed_value_set;
reg  [PRBS_WIDTH-1:0] pattern_prbs/* synthesis keep */;
wire [PRBS_WIDTH-1:0] pattern_prbs_next/* synthesis keep */;
reg  [PRBS_WIDTH-1:0] pattern_walking_bit_0/* synthesis keep */;
reg  [PRBS_WIDTH-1:0] pattern_walking_bit_1/* synthesis keep */;
reg  [PRBS_WIDTH-1:0] pattern_input_expected/* synthesis keep */;
wire [DATA_WIDTH-1:0] pattern_input_expected_all_words/* synthesis keep */;
wire                  pipeline_enable;
reg                   pipeline_enable_d1;
reg                   pipeline_enable_d2;
wire [DATA_WIDTH-1:0] data_in/* synthesis keep */;
reg  [DATA_WIDTH-1:0] data_in_d1;
wire                  decrement_payload_length_counter/* synthesis keep */;
wire                  data_compare/* synthesis keep */;
reg                   data_compare_d1;

//---------------------------------------------------------------------------------------
// CSR interface
//---------------------------------------------------------------------------------------
//CSR write control
always @(posedge clk or posedge reset)
begin
  if(reset == 1'b1)
    begin
    start                   <= 0;
    seed_checker            <= 0;
    stop_on_failure         <= 0;
    infinite_payload_length <= 0;
    error_counter_clr       <= 0;
    payload_length_counter  <= 0;
    test_pattern            <= 0;
    test_constant           <= TEST_CONSTANT_DEFAULT;
    polynomial              <= POLYNOMIAL_RESET_VALUE;
    seed_value_set          <= SEED_VALUE;
    end
  else if(clear_start == 1'b1)
    begin
    start <= 0;
    end
  else if(csr_write == 1'b1)
    begin
      case(csr_address)
      5'b00000: begin
                payload_length_counter[7:0]   <= (csr_byteenable[0] == 1'b1) ? csr_writedata[7:0]   : payload_length_counter[7:0];
                payload_length_counter[15:8]  <= (csr_byteenable[1] == 1'b1) ? csr_writedata[15:8]  : payload_length_counter[15:8];
                payload_length_counter[23:16] <= (csr_byteenable[2] == 1'b1) ? csr_writedata[23:16] : payload_length_counter[23:16];
                payload_length_counter[31:24] <= (csr_byteenable[3] == 1'b1) ? csr_writedata[31:24] : payload_length_counter[31:24];
                end
      5'b00001: begin
                payload_length_counter[39:32] <= (csr_byteenable[0] == 1'b1) ? csr_writedata[7:0]   : payload_length_counter[39:32];
                payload_length_counter[47:40] <= (csr_byteenable[1] == 1'b1) ? csr_writedata[15:8]  : payload_length_counter[47:40];
                payload_length_counter[55:48] <= (csr_byteenable[2] == 1'b1) ? csr_writedata[23:16] : payload_length_counter[55:48];
                payload_length_counter[63:56] <= (csr_byteenable[3] == 1'b1) ? csr_writedata[31:24] : payload_length_counter[63:56];
                end
      5'b00010: begin
                infinite_payload_length <= (csr_byteenable[0] == 1'b1) ? csr_writedata[0]    : infinite_payload_length;
                stop_on_failure         <= (csr_byteenable[0] == 1'b1) ? csr_writedata[1]    : stop_on_failure;
                seed_checker            <= (csr_byteenable[0] == 1'b1) ? csr_writedata[2]    : seed_checker;
                start                   <= (csr_byteenable[0] == 1'b1) ? csr_writedata[3]    : start;
                error_counter_clr       <= (csr_byteenable[0] == 1'b1) ? csr_writedata[4]    : error_counter_clr;
                test_pattern            <= (csr_byteenable[1] == 1'b1) ? csr_writedata[11:8] : test_pattern;
                end
      5'b00011: begin
                start                   <= start;
                seed_checker            <= seed_checker;
                stop_on_failure         <= stop_on_failure;
                error_counter_clr       <= error_counter_clr;
                infinite_payload_length <= infinite_payload_length;
                payload_length_counter  <= payload_length_counter;
                test_pattern            <= test_pattern;
                test_constant           <= test_constant;
                polynomial              <= polynomial;
                seed_value_set          <= seed_value_set;
                end
      5'b00100: begin
                polynomial[7:0]   <= (csr_byteenable[0] == 1'b1) ? csr_writedata[7:0]   : polynomial[7:0];
                polynomial[15:8]  <= (csr_byteenable[1] == 1'b1) ? csr_writedata[15:8]  : polynomial[15:8];
                polynomial[23:16] <= (csr_byteenable[2] == 1'b1) ? csr_writedata[23:16] : polynomial[23:16];
                polynomial[31:24] <= (csr_byteenable[3] == 1'b1) ? csr_writedata[31:24] : polynomial[31:24];
                end
      5'b00101: begin
                polynomial[39:32] <= (csr_byteenable[0] == 1'b1) ? csr_writedata[7:0]   : polynomial[39:32];
                polynomial[47:40] <= (csr_byteenable[1] == 1'b1) ? csr_writedata[15:8]  : polynomial[47:40];
                polynomial[55:48] <= (csr_byteenable[2] == 1'b1) ? csr_writedata[23:16] : polynomial[55:48];
                polynomial[63:56] <= (csr_byteenable[3] == 1'b1) ? csr_writedata[31:24] : polynomial[63:56];
                end
      5'b00110: begin
                polynomial[71:64] <= (csr_byteenable[0] == 1'b1) ? csr_writedata[7:0]   : polynomial[71:64];
                polynomial[79:72] <= (csr_byteenable[1] == 1'b1) ? csr_writedata[15:8]  : polynomial[79:72];
                polynomial[87:80] <= (csr_byteenable[2] == 1'b1) ? csr_writedata[23:16] : polynomial[87:80];
                polynomial[95:88] <= (csr_byteenable[3] == 1'b1) ? csr_writedata[31:24] : polynomial[95:88];
                end
      5'b00111: begin
                polynomial[103:96]  <= (csr_byteenable[0] == 1'b1) ? csr_writedata[7:0]   : polynomial[103:96];
                polynomial[111:104] <= (csr_byteenable[1] == 1'b1) ? csr_writedata[15:8]  : polynomial[111:104];
                polynomial[119:112] <= (csr_byteenable[2] == 1'b1) ? csr_writedata[23:16] : polynomial[119:112];
                polynomial[127:120] <= (csr_byteenable[3] == 1'b1) ? csr_writedata[31:24] : polynomial[127:120];
                end
      5'b01000: begin
                start                   <= start;
                seed_checker            <= seed_checker;
                stop_on_failure         <= stop_on_failure;
                error_counter_clr       <= error_counter_clr;
                infinite_payload_length <= infinite_payload_length;
                payload_length_counter  <= payload_length_counter;
                test_pattern            <= test_pattern;
                test_constant           <= test_constant;
                polynomial              <= polynomial;
                seed_value_set          <= seed_value_set;
                end
      5'b01001: begin
                seed_value_set[7:0]   <= (csr_byteenable[0] == 1'b1) ? csr_writedata[7:0]   : seed_value_set[7:0];
                seed_value_set[15:8]  <= (csr_byteenable[1] == 1'b1) ? csr_writedata[15:8]  : seed_value_set[15:8];
                seed_value_set[23:16] <= (csr_byteenable[2] == 1'b1) ? csr_writedata[23:16] : seed_value_set[23:16];
                seed_value_set[31:24] <= (csr_byteenable[3] == 1'b1) ? csr_writedata[31:24] : seed_value_set[31:24];
                end
      5'b01010: begin
                seed_value_set[39:32] <= (csr_byteenable[0] == 1'b1) ? csr_writedata[7:0]   : seed_value_set[39:32];
                seed_value_set[47:40] <= (csr_byteenable[1] == 1'b1) ? csr_writedata[15:8]  : seed_value_set[47:40];
                seed_value_set[55:48] <= (csr_byteenable[2] == 1'b1) ? csr_writedata[23:16] : seed_value_set[55:48];
                seed_value_set[63:56] <= (csr_byteenable[3] == 1'b1) ? csr_writedata[31:24] : seed_value_set[63:56];
                end
      5'b01011: begin
                seed_value_set[71:64] <= (csr_byteenable[0] == 1'b1) ? csr_writedata[7:0]   : seed_value_set[71:64];
                seed_value_set[79:72] <= (csr_byteenable[1] == 1'b1) ? csr_writedata[15:8]  : seed_value_set[79:72];
                seed_value_set[87:80] <= (csr_byteenable[2] == 1'b1) ? csr_writedata[23:16] : seed_value_set[87:80];
                seed_value_set[95:88] <= (csr_byteenable[3] == 1'b1) ? csr_writedata[31:24] : seed_value_set[95:88];
                end
      5'b01100: begin
                seed_value_set[103:96]  <= (csr_byteenable[0] == 1'b1) ? csr_writedata[7:0]   : seed_value_set[103:96];
                seed_value_set[111:104] <= (csr_byteenable[1] == 1'b1) ? csr_writedata[15:8]  : seed_value_set[111:104];
                seed_value_set[119:112] <= (csr_byteenable[2] == 1'b1) ? csr_writedata[23:16] : seed_value_set[119:112];
                seed_value_set[127:120] <= (csr_byteenable[3] == 1'b1) ? csr_writedata[31:24] : seed_value_set[127:120];
                end
      5'b01101: begin
                test_constant[7:0]   <= (csr_byteenable[0] == 1'b1) ? csr_writedata[7:0]   : test_constant[7:0];
                test_constant[15:8]  <= (csr_byteenable[1] == 1'b1) ? csr_writedata[15:8]  : test_constant[15:8];
                test_constant[23:16] <= (csr_byteenable[2] == 1'b1) ? csr_writedata[23:16] : test_constant[23:16];
                test_constant[31:24] <= (csr_byteenable[3] == 1'b1) ? csr_writedata[31:24] : test_constant[31:24];
                end
      5'b01110: begin
                test_constant[39:32] <= (csr_byteenable[0] == 1'b1) ? csr_writedata[7:0]   : test_constant[39:32];
                test_constant[47:40] <= (csr_byteenable[1] == 1'b1) ? csr_writedata[15:8]  : test_constant[47:40];
                test_constant[55:48] <= (csr_byteenable[2] == 1'b1) ? csr_writedata[23:16] : test_constant[55:48];
                test_constant[63:56] <= (csr_byteenable[3] == 1'b1) ? csr_writedata[31:24] : test_constant[63:56];
                end
      5'b01111: begin
                test_constant[71:64] <= (csr_byteenable[0] == 1'b1) ? csr_writedata[7:0]   : test_constant[71:64];
                test_constant[79:72] <= (csr_byteenable[1] == 1'b1) ? csr_writedata[15:8]  : test_constant[79:72];
                test_constant[87:80] <= (csr_byteenable[2] == 1'b1) ? csr_writedata[23:16] : test_constant[87:80];
                test_constant[95:88] <= (csr_byteenable[3] == 1'b1) ? csr_writedata[31:24] : test_constant[95:88];
                end
      5'b10000: begin
                test_constant[103:96]  <= (csr_byteenable[0] == 1'b1) ? csr_writedata[7:0]   : test_constant[103:96];
                test_constant[111:104] <= (csr_byteenable[1] == 1'b1) ? csr_writedata[15:8]  : test_constant[111:104];
                test_constant[119:112] <= (csr_byteenable[2] == 1'b1) ? csr_writedata[23:16] : test_constant[119:112];
                test_constant[127:120] <= (csr_byteenable[3] == 1'b1) ? csr_writedata[31:24] : test_constant[127:120];
                end
      5'b10001,
      5'b10010,
      5'b10011,
      5'b10100,
      5'b10101,
      5'b10110,
      5'b10111,
      5'b11000,
      5'b11001,
      5'b11010,
      5'b11011,
      5'b11100,
      5'b11101,
      5'b11110,
      5'b11111: begin
                start                   <= start;
                seed_checker            <= seed_checker;
                stop_on_failure         <= stop_on_failure;
                error_counter_clr       <= error_counter_clr;
                infinite_payload_length <= infinite_payload_length;
                payload_length_counter  <= payload_length_counter;
                test_pattern            <= test_pattern;
                test_constant           <= test_constant;
                polynomial              <= polynomial;
                seed_value_set          <= seed_value_set;
                end
      default:  begin
                start                   <= 0;
                seed_checker            <= 0;
                stop_on_failure         <= 0;
                error_counter_clr       <= 0;
                infinite_payload_length <= 0;
                payload_length_counter  <= 0;
                test_pattern            <= 0;
                test_constant           <= TEST_CONSTANT_DEFAULT;
                polynomial              <= POLYNOMIAL_RESET_VALUE;
                seed_value_set          <= SEED_VALUE;
                end
      endcase
    end
  else if(decrement_payload_length_counter == 1'b1)
    begin
    error_counter_clr      <= 0;
    payload_length_counter <= payload_length_counter - NUM_OF_SYMBOLS;
    end
  else
    begin
    start                   <= start;
    payload_length_counter  <= payload_length_counter;
    infinite_payload_length <= infinite_payload_length;
    stop_on_failure         <= stop_on_failure;
    seed_checker            <= seed_checker;
    error_counter_clr       <= 0;
    test_pattern            <= test_pattern;
    test_constant           <= test_constant;
    polynomial              <= polynomial;
    seed_value_set          <= seed_value_set;
    end
end

//CSR read control
always @(posedge clk or posedge reset)
begin
  if(reset == 1'b1)
    csr_readdata <= 0;
   else if(csr_read == 1'b1)
     begin
       case(csr_address)
       5'b00000: csr_readdata <= payload_length_counter[31:0];
       5'b00001: csr_readdata <= payload_length_counter[63:32];
       5'b00010: csr_readdata <= {20'h00000,                // [31:12]: reserved
                                  test_pattern,             // [11:8]:  test pattern
                                  2'b00,                    // [7:6]:   reserved
                                  failure_detected,         // [5]:     failure detected
                                  error_counter_clr,        // [4]:     error counter clearance
                                  start,                    // [3]:     start/stop
                                  seed_checker,             // [2]:     seed checker reload
                                  stop_on_failure,          // [1]:     stop on failure
                                  infinite_payload_length}; // [0]:     infinite test mode
       5'b00011: csr_readdata <= 0; // reserved
       5'b00100: csr_readdata <= polynomial[31:0];
       5'b00101: csr_readdata <= polynomial[63:32];
       5'b00110: csr_readdata <= polynomial[95:64];
       5'b00111: csr_readdata <= polynomial[127:96];
       5'b01000: csr_readdata <= error_counter;
       5'b01001: csr_readdata <= seed_value_set[31:0];
       5'b01010: csr_readdata <= seed_value_set[63:32];
       5'b01011: csr_readdata <= seed_value_set[95:64];
       5'b01100: csr_readdata <= seed_value_set[127:96];
       5'b01101: csr_readdata <= test_constant[31:0];
       5'b01110: csr_readdata <= test_constant[63:32];
       5'b01111: csr_readdata <= test_constant[95:64];
       5'b10000: csr_readdata <= test_constant[127:96];
       5'b10001,
       5'b10010,
       5'b10011,
       5'b10100,
       5'b10101,
       5'b10110,
       5'b10111,
       5'b11000,
       5'b11001,
       5'b11010,
       5'b11011,
       5'b11100,
       5'b11101,
       5'b11110,
       5'b11111: csr_readdata <= 0; // reserved
       default:  csr_readdata <= 0;
       endcase
     end
   else
     csr_readdata <= csr_readdata;
end

//---------------------------------------------------------------------------------------
// Pattern checker
//---------------------------------------------------------------------------------------
always @(posedge clk or posedge reset)
begin
  if(reset == 1'b1)
    start_d1 <= 0;
  else
    start_d1 <= start;
end

//pattern check enable control
assign pipeline_enable = (start == 1) & (snk_valid == 1);

always @(posedge clk or posedge reset)
begin
  if(reset == 1'b1)
    begin
    pipeline_enable_d1 <= 0;
    pipeline_enable_d2 <= 0;
    end
  else
    begin
    pipeline_enable_d1 <= pipeline_enable;
    pipeline_enable_d2 <= pipeline_enable_d1;
    end
end

//error statistics
always @(posedge clk or posedge reset)
begin
  if(reset == 1'b1)
    error_counter <= 0;
  else if(error_counter_clr == 1'b1)
    error_counter <= 0;
  else if(data_compare == 1'b0)
    error_counter <= error_counter + 1;
  else
    error_counter <= error_counter;
end

//failure detection
always @(posedge clk or posedge reset)
begin
  if(reset == 1'b1)
    data_compare_d1 <= 0;
  else if (pipeline_enable_d1 == 1'b1)
    data_compare_d1 <= data_compare;
  else
    data_compare_d1 <= data_compare_d1;
end

//clear right after failure happens, just for stop-on-failure trigger
always @(posedge clk or posedge reset)
begin
  if(reset == 1'b1)
    failure_detected <= 1'b0;
  else if((start == 1'b1) && (start_d1 == 1'b0)) // cleared every time right after next round test starts
    failure_detected <= 1'b0;
  else if(failure_detected_set == 1'b1)
    failure_detected <= 1'b1;
  else
    failure_detected <= failure_detected;
end

//----------------------------------------------------------------------------
// PRBS generation
//----------------------------------------------------------------------------
always @(posedge clk or posedge reset)
begin
  if(reset == 1'b1)
    pattern_prbs <= SEED_VALUE;
  else if(((start == 1'b1) && (start_d1 == 1'b0)) && (seed_checker == 1'b1))
    pattern_prbs <= seed_value_set[PRBS_WIDTH-1:0];
  else if(pipeline_enable == 1'b1)
    pattern_prbs <= pattern_prbs_next;
  else
    pattern_prbs <= pattern_prbs;
end

//LSFR input data for all the pattern_prbs stages
genvar d;
generate
  for (d = 0; d < (PRBS_WIDTH-1); d = d + 1) // highest order pattern_prbs input will be generated independent of this loop 
  begin: pipeline_taps
    assign pattern_prbs_next[d] = pattern_prbs[d+1] ^ (pattern_prbs[0] & polynomial[d]);
  end
endgenerate

assign pattern_prbs_next[PRBS_WIDTH-1] = pattern_prbs[0];

//----------------------------------------------------------------------------
// Walking '0' or '1' data pattern generation/control
//----------------------------------------------------------------------------
always @(posedge clk or posedge reset)
begin
  if(reset == 1'b1)
    pattern_walking_bit_0 <= PATTERN_WALKING_0;
  else if(((start == 1'b1) && (start_d1 == 1'b0)) && (seed_checker == 1'b1))
    pattern_walking_bit_0 <= PATTERN_WALKING_0;
  else if(pipeline_enable == 1'b1)
    pattern_walking_bit_0 <= {pattern_walking_bit_0[PRBS_WIDTH-2:0], pattern_walking_bit_0[PRBS_WIDTH-1]};
  else
    pattern_walking_bit_0 <= pattern_walking_bit_0;
end

always @(posedge clk or posedge reset)
begin
  if(reset == 1'b1)
    pattern_walking_bit_1 <= PATTERN_WALKING_1;
  else if(((start == 1'b1) && (start_d1 == 1'b0)) && (seed_checker == 1'b1))
    pattern_walking_bit_1 <= PATTERN_WALKING_1;
  else if(pipeline_enable == 1'b1)
    pattern_walking_bit_1 <= {pattern_walking_bit_1[PRBS_WIDTH-2:0], pattern_walking_bit_1[PRBS_WIDTH-1]};
  else
    pattern_walking_bit_1 <= pattern_walking_bit_1;
end

//----------------------------------------------------------------------------
// data pattern expected control
always @(posedge clk or posedge reset)
begin
  if(reset == 1'b1)
    pattern_input_expected <= 0;
  else if(pipeline_enable == 1'b1)
    begin
      if(test_pattern == 3'b001)
        pattern_input_expected <= test_constant[PRBS_WIDTH-1:0];
      else if(test_pattern == 3'b010)
        pattern_input_expected <= pattern_walking_bit_0;
      else if(test_pattern == 3'b011)
        pattern_input_expected <= pattern_walking_bit_1;
      else
        pattern_input_expected <= pattern_prbs;
    end
  else
    pattern_input_expected <= pattern_input_expected;
end

//----------------------------------------------------------------------------
// data pattern input control
always @(posedge clk or posedge reset)
begin
  if(reset == 1'b1)
    data_in_d1 <= 0;
  else if (pipeline_enable == 1'b1)
    data_in_d1 <= data_in;
  else
    data_in_d1 <= data_in_d1;
end

assign decrement_payload_length_counter = (start == 1) & (snk_valid == 1) & (infinite_payload_length == 0);
assign clear_start = (start == 1) & (((failure_detected == 1) & (stop_on_failure == 1)) | ((infinite_payload_length == 0) & (payload_length_counter == 0)));

assign pattern_input_expected_all_words = {(DATA_WIDTH/PRBS_WIDTH){pattern_input_expected}};
assign data_compare = (data_in_d1 == pattern_input_expected_all_words) | (!pipeline_enable_d1);
assign failure_detected_set = (data_compare_d1 == 0) & (pipeline_enable_d2 == 1);

assign snk_ready = start;
assign data_in   = snk_data;

endmodule